/*******************************************************************************
 * Copyright (c) 2012 Simon Urli.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Simon Urli - initial API and implementation
 ******************************************************************************/
package fr.unice.i3s.modalis.yourcast.multipleSPL;

import java.util.*;

import fr.unice.i3s.modalis.yourcast.featureModel.FeatureModel;
import fr.unice.i3s.modalis.yourcast.multiVariant.Context;

/**
 * A Composite Software Product Line (CSPL) contains :<br />
 * 		- a unique global context	<br />
 * 		- a list of local contexts	<br />
 * 		- a map of feature models and their associations	<br />
 * 		- a map of feature models and their cardinalities
 * @author urli
 *
 */
public class CSPL {

	private List<Context> contexts;
	private Map<String,List<FMAssociation>> associations;
	private Map<String,Cardinality> fmsCards;
	private Context global;
	
	public CSPL(Map<String,List<FMAssociation>> associations, Map<FeatureModel,Cardinality> fmsCards) {
		this.contexts = new ArrayList<Context>();
		this.associations = associations;
		this.fmsCards = new HashMap<String,Cardinality>();
		
		// build the first set of restricted feature models: just a clone of feature models in input
		// the global context is created from this set
		Map<String,FeatureModel> restrictedFM = new HashMap<String,FeatureModel>();
		FeatureModel f;
		for (FeatureModel fm : fmsCards.keySet()) {
			this.fmsCards.put(fm.getName(), fmsCards.get(fm));
			f = fm.clone();
			restrictedFM.put(f.getName(),f);
		}
		this.global = new Context(restrictedFM);
	}
	
	public void addContext(Context c) {
		this.contexts.add(c);
	}
	
	/**
	 * @return the multiplicity of feature models fM.
	 */
	public Cardinality mult(FeatureModel fM) {
		return this.fmsCards.get(fM.getName());
	}

	public FeatureModel getRestrictedFMFromGlobalContext(FeatureModel f) {
		return this.global.getFM(f);
	}

	public void putInGlobalContext(FeatureModel n) {
		this.global.put(n);
	}

	public FeatureModel getFM(String target) {
		return this.global.getFMByName(target);
	}
	
	public Context getGlobalContext() {
		return this.global;
	}
	
	public List<Context> getAllContexts() {
		return this.contexts;
	}
	
	public void modifyAllContextsWith(FeatureModel n) {
		for (Context c : this.contexts)
			c.put(n);
	}
	
	public List<FMAssociation> getAssociations(FeatureModel fm) {
		return this.associations.get(fm.getName());
	}
	
	public static void propagate(CSPL cspl, FeatureModel FM, Context context) {
		List<FeatureModel> FMs = propagateSpecialization(cspl, FM);
		List<FeatureModel> FMg = new ArrayList<FeatureModel>();
		List<FeatureModel> FMl = new ArrayList<FeatureModel>();
		separate(cspl,FMs,FMg,FMl);
		localPropagation(context,FMl);
		List<Context> Cls = cspl.getAllContexts();
		Context Cg = cspl.getGlobalContext();
		globalPropagation(Cls, Cg, FMg);
	}
	
	private static List<FeatureModel> propagateSpecialization(CSPL cspl, FeatureModel FM) {
		Context virtualContext = new Context(FM);
		propagateSpecializationKeepingVisited(cspl, FM, virtualContext);
		return virtualContext.getAllFMs();
	}
	
	private static void propagateSpecializationKeepingVisited(CSPL cspl, FeatureModel FM, Context virtualContext) {
		for (FMAssociation r : cspl.getAssociations(FM)) {
			FeatureModel target = r.getTargetFM(cspl);
			if (virtualContext.containsFM(target.getName()))
				target = virtualContext.getFM(target);
			FeatureModel f = r.restrict(target, FM);

			if (!f.equals(target)) {
				virtualContext.put(f);
				propagateSpecializationKeepingVisited(cspl, f, virtualContext);
			}
		}
	}
	
	private static void separate(CSPL cspl, List<FeatureModel> allFMs, List<FeatureModel> globalFMs, List<FeatureModel> localFMs) {
		for (FeatureModel fm : allFMs) {
			if (cspl.mult(fm) == Cardinality.MANY)
				localFMs.add(fm);
			else
				globalFMs.add(fm);
		}
	}
	
	private static void localPropagation(Context context, List<FeatureModel> fms) {
		for (FeatureModel fm : fms) {	
			context.put(fm);
		}
	}
	
	private static void globalPropagation(List<Context> localContexts, Context globalContext, List<FeatureModel> fms) {
		localPropagation(globalContext, fms);
		for (Context c : localContexts) {
			for (FeatureModel fm : fms) {
				if (c.containsFM(fm.getName())) {
					c.put(fm);
				}
			}
		}
	}
}
